# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import collections
from dewi.core.context import Context


class Plugin:
    """
    A plugin is an extension of DEWI.
    """

    def get_description(self) -> str:
        raise NotImplementedError

    def get_dependencies(self) -> collections.Iterable:
        return ()

    def load(self, c: Context):
        raise NotImplementedError



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import argparse
import collections
from dewi.core.command import Command
from dewi.core.context import Context
from dewi.loader.plugin import Plugin


class SampleCommand(Command):
    name = 'sample'

    def run(self, ns: argparse.ArgumentParser):
        return 42


class SamplePlugin(Plugin):
    def load(self, c: Context):
        c['commands'].register_class(SampleCommand)

    def get_description(self) -> str:
        return "plugin description"

    def get_dependencies(self) -> collections.Iterable:
        return {'dewi.core.CorePlugin'}



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3
import collections


class ContextError(Exception):
    pass


class ContextEntryNotFound(ContextError):
    pass


class ContextEntryAlreadyRegistered(ContextError):
    pass


class Context(collections.Mapping):
    """
    A context is a generic purpose registry, which helps
    communicate between different parts of the code. Instead of a global variable
    or module is for storing object(s), the context can be used and passed around
    the necessary objects and functions.

    An example: a CommandRegistry object can be registered into this context.
    """
    def __init__(self):
        self.__entries = {}

    def register(self, name: str, value):
        """
        Registers an element into the context. It doesn't support overwriting already
        registered entries, in that case `ContextEntryAlreadyRegistered` will be raised.

        :param name: the name of the new entry
        :param value: the value of the new entry
       """
        if name in self.__entries:
            raise ContextEntryAlreadyRegistered("Context entry {!r} already registered".format(name))
        self.__entries[name] = value

    def unregister(self, name: str):
        """
        Unregisters an already registered entry
        :param name: The name of the entry to be unregistered
        """
        self.__check_entry(name)

        del self.__entries[name]

    def __check_entry(self, name: str):
        if name not in self.__entries:
            raise ContextEntryNotFound("Requested context entry {!r} is not registered".format(name))

    def __len__(self):
        return len(self.__entries)

    def __getitem__(self, item):
        self.__check_entry(item)

        return self.__entries[item]

    def __contains__(self, item):
        return item in self.__entries

    def __iter__(self):
        return iter(self.__entries)



# Copyright 2012-2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import collections
import sys

from dewi.core.command import Command
from dewi.core.context import Context
from dewi.loader.plugin import Plugin


class CommandRegistryException(Exception):
    pass


class ClassNotFound(CommandRegistryException):
    pass


class ClassIsNotSubclassOfCommand(Exception):
    pass


class ClassDescriptor(object):
    """
    Describes how to create a class based on specific values in subclasses.

    There is also a method that lets the class to be get - without instantiation of it.
    """

    def get_class(self):
        raise NotImplementedError()

    def get_name(self):
        raise NotImplementedError()


class ClassDescriptorWithModuleAndClassName(ClassDescriptor):
    """
    Both the module name and the command class name are specified
    """
    def __init__(self, module_name: str, class_name: str):
        super().__init__()
        self._module_name = module_name
        self._class_name = class_name

    def get_class(self) -> Command:
        module = self._get_module_object()
        try:
            return getattr(module, self._class_name)
        except AttributeError:
            raise ClassNotFound("Invalid class or attribute: {0}".format(self._class_name))

    def _get_module_object(self):
        __import__(self._module_name)
        return sys.modules[self._module_name]

    def get_name(self) -> str:
        return self.get_class().name


class ClassDescriptorWithModuleName(ClassDescriptorWithModuleAndClassName):
    """The module and the class has the same name (the module's last part at least)"""

    def __init__(self, module_name: str):
        super().__init__(module_name, module_name.split('.')[-1])

    def get_name(self) -> str:
        return self.get_class().name


class ClassDescriptorWithModuleNameAndCommandClassMember(ClassDescriptorWithModuleAndClassName):
    """
    The command is in a module that have specific layout, its __init__.py has commandlcass member. This member
    stores the command class.
    """
    def __init__(self, module_name: str):
        super().__init__(module_name, 'command_class')


class ClassDescriptorWithConcreteClass(ClassDescriptor):
    """
    This descriptor stores the class itself, mainly used for testing.
    """
    def __init__(self, class_object: type):
        if not issubclass(class_object, Command):
            raise ClassIsNotSubclassOfCommand(
                    'The {} class is not subclass of dewi.core.command.Command'.format(class_object.__name__))
        super().__init__()
        self.class_object = class_object

    def get_class(self) -> Command:
        return self.class_object

    def get_name(self) -> str:
        return self.class_object.name


class CommandRegistry(object):
    '''
    Registry of command classes
    '''

    def __init__(self):
        self.__registry = dict()

    def __validate_name_and_class_descriptor(self, name: str, class_descriptor: ClassDescriptor):
        if name in self.__registry:
            raise CommandRegistryException("Command class name is already used; name='{0}'".format(name))
        if not isinstance(class_descriptor, ClassDescriptor):
            raise CommandRegistryException(
                "Command class descriptor's type differs from ClassDescriptor; "
                "type='{0}'".format(type(class_descriptor).__name__))

    def register_command_class(self, name: str, class_descriptor: ClassDescriptor):
        self.__validate_name_and_class_descriptor(name, class_descriptor)
        self.__registry[name] = class_descriptor

    def get_command_class_descriptor(self, name: str) -> ClassDescriptor:
        if name not in self.__registry:
            raise CommandRegistryException("Specified command class name is not found; name='{0}'".format(name))
        return self.__registry[name]

    def get_command_count(self) -> int:
        return len(self.__registry)

    def get_command_names(self) -> collections.Sequence:
        return list(self.__registry.keys())


class CommandRegistrar:
    """
    Wrap a CommandRegistry object for easier registration
    """

    def __init__(self, cr: CommandRegistry):
        self.__registry = cr

    def register_class(self, command_class: type):
        desc = ClassDescriptorWithConcreteClass(command_class)
        self.__registry.register_command_class(command_class.name, desc)
        for alias in command_class.aliases:
            self.__registry.register_command_class(alias, desc)


class CommandRegistryPlugin(Plugin):
    def get_description(self):
        return "Registers command-related entries into context"

    def load(self, c: Context):
        c.register('commandregistry', CommandRegistry())
        c.register('commands', CommandRegistrar(c['commandregistry']))



import io
import unittest
import sys


def redirect_outputs(stdout=None, stderr=None):
    class Redirection:
        def __init__(self):
            self.stdout = stdout or io.StringIO()
            self.stderr = stderr or io.StringIO()

            self.__saved_stdouts = []
            self.__saved_stderrs = []

        def __enter__(self):
            self.__saved_stdouts.append(sys.stdout)
            self.__saved_stderrs.append(sys.stderr)

            sys.stdout = self.stdout
            sys.stderr = self.stderr

            return self

        def __exit__(self, *args):
            sys.stdout = self.__saved_stdouts.pop()
            sys.stderr = self.__saved_stderrs.pop()

    return Redirection()


class TestCase(unittest.TestCase):
    maxDiff = None

    def set_up(self):
        pass

    def tear_down(self):
        pass

    def setUp(self):
        self.set_up()

    def tearDown(self):
        self.tear_down()

    assert_almost_equal = unittest.TestCase.assertAlmostEqual
    assert_count_equal = unittest.TestCase.assertCountEqual
    assert_dict_equal = unittest.TestCase.assertDictEqual
    assert_equal = unittest.TestCase.assertEqual
    assert_false = unittest.TestCase.assertFalse
    assert_greater = unittest.TestCase.assertGreater
    assert_greater_equal = unittest.TestCase.assertGreaterEqual
    assert_in = unittest.TestCase.assertIn
    assert_is = unittest.TestCase.assertIs
    assert_is_instance = unittest.TestCase.assertIsInstance
    assert_is_none = unittest.TestCase.assertIsNone
    assert_is_not = unittest.TestCase.assertIsNot
    assert_is_not_none = unittest.TestCase.assertIsNotNone
    assert_less = unittest.TestCase.assertLess
    assert_less_equal = unittest.TestCase.assertLessEqual
    assert_list_equal = unittest.TestCase.assertListEqual
    assert_logs = unittest.TestCase.assertLogs
    assert_multi_line_equal = unittest.TestCase.assertMultiLineEqual
    assert_not_almost_equal = unittest.TestCase.assertNotAlmostEqual
    assert_not_equal = unittest.TestCase.assertNotEqual
    assert_not_in = unittest.TestCase.assertNotIn
    assert_not_is_instance = unittest.TestCase.assertNotIsInstance
    assert_not_regex = unittest.TestCase.assertNotRegex
    assert_raises = unittest.TestCase.assertRaises
    assert_raises_regex = unittest.TestCase.assertRaisesRegex
    assert_regex = unittest.TestCase.assertRegex
    assert_sequence_equal = unittest.TestCase.assertSequenceEqual
    assert_set_equal = unittest.TestCase.assertSetEqual
    assert_true = unittest.TestCase.assertTrue
    assert_tuple_equal = unittest.TestCase.assertTupleEqual
    assert_warns = unittest.TestCase.assertWarns
    assert_warns_regex = unittest.TestCase.assertWarnsRegex



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3


import sys
from dewi.core.application import MainApplication
from dewi.loader.loader import PluginLoader


def main():
    loader = PluginLoader()
    app = MainApplication(loader, 'dewi')
    app.run(sys.argv[1:])

if __name__ == '__main__':
    main()



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import collections

from dewi.core.commandregistry import CommandRegistry, CommandRegistrar
from dewi.core.context import Context

from dewi.loader.plugin import Plugin


class CorePlugin(Plugin):
    def get_description(self):
        return "Core features of DEWI"

    def get_dependencies(self) -> collections.Iterable:
        return ('dewi.core.commandregistry.CommandRegistryPlugin',)

    def load(self, c: Context):
        pass



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import argparse
import collections
import sys

from dewi.core.context import Context
from dewi.loader.loader import PluginLoader
from dewi.loader.plugin import Plugin


class DewiPlugin(Plugin):
    def get_description(self) -> str:
        return "DEWI application plugin"

    def get_dependencies(self) -> collections.Iterable:
        return {'dewi.core.CorePlugin', 'dewi.commands.FileManipulationPlugin'}

    def load(self, c: Context):
        pass


class MainApplication:
    def __init__(self, loader: PluginLoader, program_name: str):
        self.__loader = loader
        self.__program_name = program_name

    def __parse_app_args(self, args: collections.Iterable):
        parser = argparse.ArgumentParser(
            prog=self.__program_name,
            usage='%(prog)s [options] [command [command-args]]')
        parser.add_argument(
            '-p', '--plugin', help='Load this plugin. This option can be specified more than once.',
            default=[], action='append')
        parser.add_argument('command', nargs=1, help='Command to be run')
        parser.add_argument(
            'commandargs', nargs=argparse.REMAINDER, help='Additonal options and arguments of the specified command',
            default=[],)
        return parser.parse_args(args)

    def run(self, args: collections.Iterable):
        app_ns = self.__parse_app_args(args)

        plugins = app_ns.plugin or ['dewi.core.application.DewiPlugin']
        plugins.append('dewi.core.commandregistry.CommandRegistryPlugin')
        try:
            context = self.__loader.load(set(plugins))
            command_name = app_ns.command[0]

            command_class = context['commandregistry'].get_command_class_descriptor(command_name).get_class()
            command = command_class()
            parser = argparse.ArgumentParser(
                description=command.description,
                prog='{} {}'.format(self.__program_name, command.name))

            command.register_arguments(parser)
            ns = parser.parse_args(app_ns.commandargs)
            sys.exit(command.run(ns))
        except SystemExit:
            raise
        except BaseException as exc:
            print(exc, file=sys.stderr)
            sys.exit(1)

        sys.exit(0)



# Copyright (c) 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import collections

from dewi.commands.edit.edit import EditCommand
from dewi.core.context import Context
from dewi.loader.plugin import Plugin


class FileManipulationPlugin(Plugin):
    def get_description(self) -> str:
        return "File manipulation commnads"

    def get_dependencies(self) -> collections.Iterable:
        return {'dewi.core.CorePlugin'}

    def load(self, c: Context):
        c['commands'].register_class(EditCommand)



# Copyright 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import argparse


class Command:
    name = ''
    aliases = list()
    description = ''

    def register_arguments(self, parser: argparse.ArgumentParser) -> None:
        pass

    def run(self, args: argparse.Namespace):
        raise NotImplementedError



# Copyright (c) 2015 Laszlo Attila Toth
# Distributed under the terms of the GNU General Public License v3

import argparse

import re
import subprocess
from dewi.core.command import Command


def convert_to_vim_args(args):
    result = []
    if len(args) == 1:
        match = re.match(r'(.*?):([[0-9]+)(:[0-9]+)?:?$', args[0])
        if match:
            result.append(match.group(1))
            result.append('+' + match.group(2))
        else:
            result.append(args[0])
    elif len(args) > 1:
        result = ['-p'] + args

    return result


class EditCommand(Command):
    name = 'edit'
    aliases = ['ed']
    description = 'Calls vim with the file names and line numbers parsed from argument list.'

    def register_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument('file_list', nargs=argparse.REMAINDER,  help='List of files for editing')

    def run(self, ns: argparse.Namespace):
        args = ['vim'] + convert_to_vim_args(ns.file_list)
        pipe = subprocess.Popen(args)
        pipe.communicate()
